/**
 * @license Copyright (c) 2012 Boris Sabadach
 * LICENSE: see the license.txt file. If file is missing, this file is subject
 * to the MIT License at: http://www.opensource.org/licenses/mit-license.php.
 */

/**
 * backbone-view-states.js
 *
 * @version 0.3.0
 * @author boris@washmatique.fr
 */
(function () {

    "use strict";

    var root = this,
        $=root.$,
        _ = root._,
        Backbone = root.Backbone,
        document = root.document,
        console = root.console,
        CSSStyleRule = root.CSSStyleRule,
        setTimeout=root.setTimeout,

        config = {
            stateAttr:"data-states",
            inAttr:"data-states-inc",
            outAttr:"data-states-exc",
            cssModifier:"-"
        },

    //---------------------------------
    // State events types
    //-----------------------------------
        EXIT = "state:exit",
        EXITED = "state:exited",
        ENTER = "state:enter",
        CHANGED = "state:changed";


    // turn off debug by default
    Backbone.DEBUG_STATES = false;


    //---------------------------------
    // utility functions
    //---------------------------------
    /**
     * debug function to log in console if available
     * @type {*}
     */
    var debug = (function () {
        var join = Array.prototype.join,
            canLog = console && console.log;
        return function () {
            return canLog && Backbone.DEBUG_STATES && console.log(join.call(arguments, ','));
        };
    }());


    /**
     * add the trim function if not availiable
     * @type {*}
     */
    if (!_.isFunction(String.prototype.trim)) {
        String.prototype.trim = function () {
            return this.replace(/^\s+|\s+$/g, "");
        };
    }


    /**
     * CSS utility functions.
     * @type {*}
     */
    var CSS = (function (document) {
        var _hasSheets = true,
            _stylesCache = {},
            _transitionsCache = {},
            _acceptedMediaTypes = ["screen", "all", ""];

        /**
         * return true if among the comma separated string list of media types declared on a sheet and passed as parameter
         * should be accepted for further process
         * @param mediaTypes
         * @return {Boolean}
         * @private
         */
        var _acceptSheet = function (sheet) {
            var mediaTypes = sheet.media.mediaText;
            mediaTypes = mediaTypes.split(",");
            return _.find(mediaTypes, function (mediaType) {
                return _.indexOf(_acceptedMediaTypes, mediaType.trim()) >= 0;
            }) !== void 0;
        };

        /**
         *  returns true if in the rule.selectorText passed as parameter one in the list (comma separated) matches
         *  the selectorText passed as parameter
         * @param rule
         * @param selectorText
         * @return {*}
         * @private
         */
        var _ruleHasSelectorText = function (rule, selectorText) {
            var selectors = rule.selectorText.split(",");
            return _.find(selectors, function (aSelectorText) {
                return aSelectorText.trim() === selectorText;
            });
        };

        /**
         * return the Style object among rules passed as parameter if one of the rules selectorText contains
         * the selector string passed a parameter.
         * @param rules a list of Rules
         * @param selector : a style selector
         * @return {*}
         * @private
         */
        var _findStyleInRules = function (rules, selector) {
            var rule;
            for (var j = 0, k = rules.length; j < k; j++) {
                rule = rules[j];
                if ((rule.constructor===CSSStyleRule) && _ruleHasSelectorText(rule, selector)) {
                    return rule.style;
                }
            }
            return null;
        };


        return {

            transitions:(function () {
                var el = document.createElement('div'),
                    transProps = ['transition', 'OTransition', 'MSTransition', 'MozTransition', 'WebkitTransition'],
                    props = el.style,
                    prop = _.find(transProps, function (value) {
                        return props[value] === void 0 ? false : value;
                    });

                return prop && {
                    durationProp:prop + "Duration",
                    delayProp:prop + "Delay"
                };
            }()),

            /**
             * returns the Style Object and caches it if a CSS rule was found among all
             * declared style sheets in the document.Sheets loaded from a different domain are excluded
             * @param selector
             * @return {*}
             */
            findStyle:function (selector) {
                if (_hasSheets && _stylesCache[selector] !== void 0) return _stylesCache[selector];
                var sheets = document.styleSheets;
                if (sheets === void 0) {
                    _hasSheets = false;
                    return null;
                }
                var sheet,
                    rules,
                    styleRule;
                for (var i = 0, l = sheets.length; i < l; i++) {
                    try {
                        sheet = sheets[i];
                        if (!sheet || !_acceptSheet(sheet)) continue; // exclude all media type except screen
                        rules = sheet.rules || sheet.cssRules;
                        if (rules === null) continue;
                        styleRule = _findStyleInRules(rules, selector);
                        if (styleRule !== null) {
                            _stylesCache[selector] = styleRule;
                            return styleRule;
                        }
                    } catch (e) {
                        console.log(e.message);
                        if (e.name === 'SecurityError') continue;
                        else throw e;
                    }
                }
                _stylesCache[selector] = "undefined";
                return _stylesCache[selector];
            },


            /**
             * returns true if a CSS rule corresponding to the selector rule was found among all
             * declared style sheets in the document
             * @param selector
             */
            hasRule:function (selector) {
                return this.findStyle(selector) !== "undefined";
            },

            /**
             * return the maximum number in milliseconds of a comma separated expression of CSS durations declarations
             * for instance '1s,2s' returns 2000.
             * @param expression
             */
            getMillis:function (expression) {
                var tokens = expression.split(","),
                    millis = 0,
                    t;
                _.each(tokens, function (token) {
                    t = parseFloat(token.replace("s", "").replace("m", ""));
                    millis = Math.max(millis, isNaN(t) ? 0 : t * (token.indexOf("ms") === -1 ? 1000 : 1));
                });
                return millis;
            },


            /**
             * return the total duration including delay of the CC3 transition in the CCS rule defined by the selector passed as parameter
             * @param selector
             */
            getTransDuration:function (selector) {
                if (!this.transitions) return 0;
                return _transitionsCache[selector] || (function (self) {
                    var style,
                        duration = 0;
                    if (self.hasRule(selector)) {
                        style = self.findStyle(selector);
                        duration = self.getMillis(style[self.transitions.durationProp]) + self.getMillis(style[self.transitions.delayProp]);
                    }
                    _transitionsCache[selector] = duration;
                    return _transitionsCache[selector];
                })(this);

            }
        };

    }(document));


    //---------
    // Core
    //---------
    /**
     * The ViewStateTarget object wraps an element (queried by the DOM selector library) that is included or excluded in a view state.
     * The class provides methods to manage the element visibility:display:block/none or visibility:visible/hidden
     * according the keepLayout option in the constructor
     * @type {*}
     */
    var ViewStateTarget = (function (CSS, config) {

        var _displayPolicies = [
            {
                type:"display",
                yes:"block",
                no:"none"
            },
            {
                type:"visibility",
                yes:"visible",
                no:"hidden"
            }

        ];

        var ViewStateTarget = function ($el, keepLayout) {
            this.$el = $el;
            this.display = _displayPolicies[keepLayout ? 1 : 0];
            this.ensure();
        };

        ViewStateTarget.prototype = {
            ensure:function () {
                if (this.$el.attr("class") === void 0) throw new Error(this.$el.selector + " element(s) must have at least a base style class");
                this.baseStyle = this.$el.attr("class").split(" ").pop();
            },

            isDisplayed:function () {
                // its' simpler to compare to the not displayed css value because in the case of display block/none
                // the initial display value of the element could not be "block"  but "inline-block" or else
                return this.$el.css(this.display.type) !== this.display.no;
            },

            toggle:function () {
                this.$el.css(this.display.type, this.isDisplayed() ? this.display.no : this.display.yes);
            },

            removeStateStyle:function (state) {
                var style = this.baseStyle + config.cssModifier + state;
                if (style && this.$el.hasClass(style)) this.$el.removeClass(style);
            },

            addStateStyle:function (state) {
                var selector = this.getStyleSelector(state);
                if (CSS.hasRule(selector)) this.$el.addClass(selector.replace(".", ""));
            },

            getStyleSelector:function (state) {
                return "." + this.baseStyle + config.cssModifier + state;
            }

        };

        return ViewStateTarget;
    }(CSS, config));


    /**
     *  The ViewState class encapsulate the different states declared on a DOM element.
     * - it references arrays of child DOM elements (state targets) that are included or excluded from the state
     * - the class provide an 'enter' and an 'exit' method that manages the state targets lifecycle during states transitions
     * @type {*}
     */
    var ViewState = (function (ViewStateTarget) {

        /**
         * @param name : the state name
         * @param $el : the DOM element targeted as a state container
         * @param keepLayout: the display policy for state targets.
         * @constructor
         */
        var ViewState = function (name, $el, keepLayout) {
            this.$el = $el;
            this.name = name;
            this.keepLayout = keepLayout;
            this.transDuration = -1;
            this.inclusions = [];
            this.exclusions = [];
        };


        ViewState.prototype = {

            addElement:function ($el, isIncluded) {
                (isIncluded ? this.inclusions : this.exclusions).push(new ViewStateTarget($el, this.keepLayout));
            },

            /**
             * the exit function removes all the state style class on each included target
             */
            exit:function () {
                _.each(this.inclusions, function (target) {
                    target.removeStateStyle(this.name);
                }, this);
            },

            /**
             * the enter state function hides all exclusions and shows all inclusions
             */
            enter:function () {
                _.each(this.exclusions, function (target) {
                    this._update(target, false);
                }, this);
                _.each(this.inclusions, function (target) {
                    this._update(target, true);
                }, this);
            },

            /**
             * update the visibility of a the state target and applies the state style class if the target has to be visible
             * @param target
             * @param mustDisplay
             * @private
             */
            _update:function (target, mustDisplay) {
                if (mustDisplay !== target.isDisplayed()) target.toggle();
                if (mustDisplay) target.addStateStyle(this.name);
            }

        };

        return ViewState;

    }(ViewStateTarget));

    /**
     * The CSS3StateTransition triggers the CSS3 transitions on a view state by applying style that contains transitions
     * and waits for the transition to be completed to invoke a callback method.
     * The listening strategy is based on full transition duration and not on transitionEnd events handling
     * @type {*}
     */

    var CSS3StateTransition = (function (CSS) {

        var CSS3StateTransition = function () {
            this.isPlaying = false;
            this.callBack = null;
            this.onCompleteCallBack = _.bind(this._onComplete, this);
        };

        CSS3StateTransition.prototype = {

            playOn:function (state) {
                this.isPlaying = true;
                this._playOn(state, false);
            },

            reverseOn:function (state) {
                this.isPlaying = true;
                this._playOn(state, true);
            },

            _playOn:function (state, reverseIt) {
                if (reverseIt) state.exit();
                else state.enter();
                var duration = this._getTransDuration(state);
                if (duration === 0) return this._onComplete();
                setTimeout(this.onCompleteCallBack, duration);
            },

            _onComplete:function () {
                this.callBack();
                this.isPlaying = false;
            },

            _getTransDuration:function (state) {
                if (state.transDuration >= 0) return state.transDuration;
                var duration = 0;
                _.each(state.inclusions, function (target) {
                    duration = Math.max(CSS.getTransDuration(target.getStyleSelector(state.name)), duration);
                });
                // cache transition in the state object
                state.transDuration = duration;
                return duration;
            }


        };

        return CSS3StateTransition;

    })(CSS);


    /**
     * the ViewStatesManager is the delegate of a StatesClient that manages its state change, without listening to transitions durations
     * @type {*}
     */
    var ViewStatesManager = (function () {

        /**
         *
         * @param attributes
         * @param options
         * @constructor
         */
        var ViewStatesManager = function (attributes, options) {
            this.view = attributes.view;
            this.states = attributes.states;
            this.initialize.call(this, options);
        };

        ViewStatesManager.prototype = {

            initialize:function (options) {
                this.silent = Boolean(options.silent);
                this.nextState = this.currentState = null;
            },

            notifyView:function (event, stateName) {
                if (!this.silent) this.view.trigger(event, event, stateName);
            },

            setState:function (name) {
                this.ensureState(name);
                this.nextState = this.states[name];
                this.exitCurrent();
                this.enterNext();
            },

            ensureState:function (name) {
                if (this.states[name] === void 0) throw new Error("unknown state:-" + name + "-");
            },

            exitCurrent:function () {
                if (!this.currentState) return;
                if (this.nextState === this.currentState.name) return;
                this.notifyView(EXIT, this.currentState.name);
                this.currentState.exit();
                this.notifyView(EXITED, this.currentState.name);
            },

            enterNext:function () {
                this.notifyView(ENTER, this.nextState.name);
                this.nextState.enter();
                this.currentState = this.nextState;
                this.afterNextEntered();
            },
            afterNextEntered:function () {
                this.currentState = this.nextState;
                this.notifyView(CHANGED, this.nextState.name);
            }
        };
        // borrows the extend function from Backbone
        ViewStatesManager.extend = Backbone.View.extend;

        return ViewStatesManager;

    }());


    /**
     * the ViewStatesTransitionsManager  manages states and CSS3 transitions on state change
     * @type {*}
     */
    var ViewStatesTransitionsManager = (function (ViewStatesManager, CSS3StateTransition) {

        var ViewStatesTransitionsManager = ViewStatesManager.extend({

            initialize:function (options) {
                ViewStatesManager.prototype.initialize.call(this, options);
                this.reverseTransitions = !!(options.reverseTransitions);
                this.transition = new CSS3StateTransition();
                this.onCurrentExited = _.bind(this.enterNext, this);
                this.afterNextEntered = _.bind(this.afterNextEntered, this);
            },

            setState:function (name) {
                if (this.transition.isPlaying) return;
                this.ensureState(name);
                this.nextState = this.states[name];
                if (this.nextState === this.currentState) return;
                this.exitCurrent();
            },

            exitCurrent:function () {
                if (!this.currentState) {
                    this.enterNext();
                    return;
                }
                this.notifyView(EXIT, this.currentState.name);
                if (this.reverseTransitions) {
                    this.transition.callBack = this.onCurrentExited;
                    this.transition.reverseOn(this.currentState);
                } else {
                    this.currentState.exit();
                    this.enterNext();
                }

            },

            enterNext:function () {
                if (this.currentState) {
                    this.notifyView(EXITED, this.currentState.name);
                }
                this.notifyView(ENTER, this.nextState.name);
                this.transition.callBack = this.afterNextEntered;
                this.transition.playOn(this.nextState);
            },

            afterNextEntered:function () {
                this.currentState = this.nextState;
                debug(this.currentState.name);
                this.notifyView(CHANGED, this.currentState.name);
            }

        });

        return ViewStatesTransitionsManager;

    }(ViewStatesManager, CSS3StateTransition));

    /**
     *  The make method returns the right state manager regarding the browser ability to supports CSS3 transitions or not.
     * @param attributes : required object members
     * @param options  : optional construction values
     * @return {*}
     */
    ViewStatesManager.make = function (attributes, options) {
        // if the browser doesn't support transitions return a simple StatesManager
        return CSS.transitions ? new ViewStatesTransitionsManager(attributes, options) : new ViewStatesManager(attributes, options);
    };


    /**
     * States builder function: search for all state targets (child elements) in a DOM element to build its view states.
     * @type {*}
     */
    var ViewStatesBuilder = (function (config, ViewState) {

        /**
         *  Create an object literal with states names as keys.
         * @param allStatesNames
         * @return {Object}
         * @private
         */
        var _createStatesObject = function (allStatesNames) {
            allStatesNames = allStatesNames.split(",");

            var statesObject = {},
                name;
            while ((name = allStatesNames.shift())) {
                name = name.trim();
                if (statesObject[name]) throw new Error("the state " + name + " is declared twice");
                statesObject[name] = null;
            }
            return statesObject;

        };


        /**
         * Add an element to a view state
         * @param states: a map of all the ViewStates
         * @param name : the name of the state on which a state target must be added
         * @param $el : the parent DOM element defined as the state container
         * @param $child : the element to be added
         * @param keepLayout : the layout policy for the child element
         * @param included : includes the state target in the View State with the "name" name if true, excludes it otherwise
         * @private
         */
        var _registerStateTarget = function (states, stateName, $el, $child, keepLayout, included) {
            stateName = stateName.trim();
            if (stateName === '') throw new Error("declared state on a child of " + $el.selector + " has an empty name");
            states[stateName] = states[stateName] || new ViewState(stateName, $el, keepLayout);
            states[stateName].addElement($child, included);
        };

        /**
         *
         * @param states
         * @param elStateNames
         * @param $container
         * @param $el
         * @param keepInLayout
         * @param includeIt
         */
        var _addElementToViewStates = function (states, elStateNames, $container, $el, keepInLayout, includeIt) {
            _.each(elStateNames, function (stateName) {
                _registerStateTarget(states, stateName, $container, $el, keepInLayout, includeIt);
            });
            _.each(_.difference(_.keys(states), elStateNames), function (stateName) {
                _registerStateTarget(states, stateName, $container, $el, keepInLayout, !includeIt);
            });
        };

        /**
         *  populate all view states with statesContext object
         * @param states
         * @param statesContext
         * @private
         */
        var _populateStates = function (states, statesContext) {
            var $el,
                declaredStates,
                included,
                $statesContainer = statesContext.container;

            _.each(statesContext.targets, function (targetProperties) {
                $el = targetProperties.$el;
                included = targetProperties.inc !== void 0;
                declaredStates = included ? targetProperties.inc : targetProperties.exc;

                if (declaredStates !== void 0) {
                    declaredStates = declaredStates.trim() === "all" ? statesContext.names : declaredStates.trim();
                    _addElementToViewStates(states, declaredStates.split(','), $statesContainer, $el, statesContext.keepLayout, included);
                }

            });
            // if includeRoot is true add the state container as an included element of every state
            if (statesContext.includeRoot) {
                _.each(statesContext.names.split(","), function (name) {
                    _registerStateTarget(states, name, $statesContainer, $statesContainer, statesContext.keepLayout, true);
                });
            }

        };


        return {
            build:function (stateContext) {
                var states = _createStatesObject(stateContext.names);
                _populateStates(states, stateContext);
                return states;
            }
        };

    }(config, ViewState));

    /**
     *
     * @type {*}
     */
    var StatesContextBuilder = (function (config, $, _) {
        /**
         * return all child elements  matching the selector value passed as parameter using a jQuery compatible selector API .
         * If deepFetch is false it returns only direct children, otherwise returns all elements.
         * @param $el
         * @param selector
         * @param deepFetch
         * @return {*}
         */
        var _findTargets = function ($el, selector, deepFetch) {
            if (deepFetch) return $el.find("[" + selector + "]");
            return $el.children("[" + selector + "]");
        };

        /**
         *  build a statesContext object from DOM data-* states declarations
         * @param $rootEl
         * @param allChildren
         * @return {Object}
         * @private
         */
        var _fromDOM = function ($statesContainer, options) {
            var targets = [],
                key,
                $elements,
                targetDef,
                $target,
                deepFetch = options.deepFetch;
            _.each([config.outAttr, config.inAttr], function (attribute) {
                $elements = _findTargets($statesContainer, attribute, deepFetch);

                key = attribute.replace(config.stateAttr + "-", "");
                _.each($elements, function (target) {
                    $target = $(target);
                    targetDef = _.find(targets, function (item) {
                        return item.$el === $target;
                    }) || {
                        $el:$target
                    };
                    targetDef[key] = $target.attr(attribute).trim();
                    targets.push(targetDef);
                });
            });
            return targets;
        };

        /**
         * build a statesContext object from states options declarations
         * @param $stateContainer
         * @param stateOptions
         * @return {Array}
         * @private
         */
        var _fromOptions = function ($stateContainer, options) {
            var targets = [],
                targetStatesDef,
                $targets,
                $target,
                allTargetOptions = options.targets;

            _.each(_.keys(allTargetOptions), function (selector) {
                $targets = $stateContainer.find(selector);
                _.each($targets, function (target) {
                    $target = $(target);
                    targetStatesDef = {
                        $el:$target
                    };
                    _.each(["inc", "exc"], function (prop) {
                        if (allTargetOptions[selector][prop] !== void 0) targetStatesDef[prop] = allTargetOptions[selector][prop];
                    });

                    if (allTargetOptions[selector].exc !== void 0) targetStatesDef.exc = allTargetOptions[selector].exc;
                    targets.push(targetStatesDef);
                });

            });
            return targets;
        };


        return {
            build:function ($statesContainer, options) {
                if (!$statesContainer || $statesContainer.size === 0) throw new Error("cannot build states on an unknown element");
                var ctx = {},
                    findTargets;
                // the state names must be declared in DOM or options
                ctx.names = options.names || $statesContainer.attr(config.stateAttr);
                if (ctx.names === void 0) throw new Error($statesContainer.selector + " has no state declared");

                ctx.container = $statesContainer;

                // apply all declared options in the context
                _.each(_.keys(options), function (prop) {
                    ctx[prop] = options[prop];
                });

                //default properties initialization
                ctx.keepLayout = ctx.keepLayout === void 0 ? true : ctx.keepLayout;
                ctx.initialState = ctx.initialState ? ctx.initialState : ctx.names.split(",")[0];

                findTargets = _.isObject(options.targets) ? _fromOptions : _fromDOM;
                ctx.targets = findTargets($statesContainer, options);
                return ctx;
            }
        };
    })(config, $, _);


    /**
     * The StateClient defines methods to enhance any Backbone view with view states feature
     * - it delegates the creation of its states to the buildStates function
     * - it delegates the management of the states lifecycle to its stateManager member
     * - it handles state life-cycle events triggered by its view member (un-directly through the stateManager) with the 'onStateEvent' function
     * @type {*}
     */
    var StateClient = (function (buildStates, createStateManager, buildStateContext, events, debug) {


        var initializeStates = function () {
            this.currentState = null;
            this.applyDefaults();
            var statesContext = buildStateContext(this.$el, this.options.states);
            var states = buildStates(statesContext);
            this.manageStates(states);
            this.setState(statesContext.initialState);
        };

        var applyDefaults = function () {
            this.options.states = this.options.states || {};
            if (this.statesDefaults !== void 0) {
                for (var prop in this.statesDefaults) {
                    if(this.statesDefaults.hasOwnProperty(prop)){
                        this.options.states[prop] = this.statesDefaults[prop];
                    }

                }
            }
        };

        var manageStates = function (states) {
            this.stateManager = createStateManager({
                view:this,
                states:states
            }, this.options.states);
            this.on(events.join(" "), this.onStateEvent, this);
        };


        var setState = function (name) {
            this.stateManager.setState(name);
            this.currentState = name;
        };

        /**
         * this function defined as the state life cycle events handler is intended to be overridden
         * @param event: the event name for instance state:enter
         * @param stateName: the state that is targeted by this event
         */
        var onStateEvent = function (event, stateName) {
            debug(this.cid + " " + this.$el.selector + " " + event + " on " + stateName);
        };


        return {
            initializeStates:initializeStates,
            applyDefaults:applyDefaults,
            manageStates:manageStates,
            setState:setState,
            onStateEvent:onStateEvent

        };

    })(ViewStatesBuilder.build, ViewStatesManager.make, StatesContextBuilder.build, [EXIT, EXITED, ENTER, CHANGED], debug);


    /**
     * The makeAsStatesClient method return a Backbone view descendant enhanced with the StateClient behaviour
     * @param View: a Backbone view or one of its descendant
     * @return a sub view of the view passed as parameter
     */
    var makeAsStatesClient = function (View) {
        // add all the functions defined in the behaviour to the view
        var StateClientView = View.extend(StateClient);

        // the render method must be overridden: view states are defined by querying the content of a DOM element attached to view.
        // If a view is rendered via a template engine, states must be defined after the render phase.
        var memo = View.prototype.render;

        StateClientView.prototype.render = function () {
            var rendered = memo.call(this);
            this.initializeStates();
            return rendered || this.$el.html();
        };

        return StateClientView;

    };


    Backbone.States = {
        ENTER:ENTER,
        EXIT:EXIT,
        EXITED:EXITED,
        CHANGE:CHANGED,
        config:config,
        // give public access for tests
        _StatesManager:ViewStatesManager,
        _StatesTransitionsManager:ViewStatesTransitionsManager,
        _ViewState:ViewState,
        _ViewStateTarget:ViewStateTarget,
        _buildStates:ViewStatesBuilder.build,
        _createStateManager:ViewStatesManager.make,
        _CSS:CSS,
        _buildStatesContext:StatesContextBuilder.build
    };

    // define and add the StatesClientView to the root of the Backbone object
    Backbone.StatesClientView = makeAsStatesClient(Backbone.View);

    // add the makeAsStatesClient as a static method of the Backbone.View class
    Backbone.View.makeAsStatesClient = makeAsStatesClient;


}).call(this);
