(function (root) {

    "use strict";

    var Backbone = root.Backbone,
        _ = root._,
        $ = root.$,
        describe = root.describe,
        expect = root.expect,
        it = root.it,
        spyOn = root.spyOn,
        CSSStyleDeclaration = root.CSSStyleDeclaration;


    Backbone.DEBUG_STATES = true;

    var createMockHtmlContainer = function () {
        var node = "<div id='container-test' class='container' data-states='one,two,three'>";
        node = node.concat("<div id='item1' class='item1' data-states-inc='one,two,three'></div>");
        node = node.concat("<div id='item2' class='item2' data-states-inc='one,two'></div>");
        node = node.concat("<div id='item3' class='item3' data-states-exc='one'><div id='item31' class='square3' data-states-inc='one'></div></div>");
        node = node.concat("</div>");
        return node;
    };


    describe("CSS introspection tests", function () {
        var CSS = Backbone.States._CSS;

        it("should find the css rule present in the stylesheet", function () {
            var selector = ".item1-one",
                expectedStyle;
            expectedStyle = CSS.findStyle(selector);
            expect(expectedStyle).not.toEqual("undefined");

            selector = ".item1--one";
            expectedStyle = CSS.findStyle(selector);
            expect(expectedStyle).not.toEqual("undefined");

            selector = ".item2-two";
            expectedStyle = CSS.findStyle(selector);
            expect(expectedStyle).not.toEqual("undefined");
        });

        it("CSS.findStyle should return a CSSStyleDeclaration object type", function () {
            var selector = ".item1-one",
                expectedStyle;
            expectedStyle = CSS.findStyle(selector);
            expect(expectedStyle instanceof  CSSStyleDeclaration).toBeTruthy();
        });

        it("should return the string 'undefined' if the css class is not found", function () {
            var style = ".item99-one",
                found;
            found = CSS.findStyle(style);
            expect(found).toEqual("undefined");

        });

        // This part of tests is executed only when the browser supports transitions
        if (Backbone.States._CSS.transitions !== false) {
            it("CSS.getTransDuration should return the proper total value of transition duration if the css class has transitions and the browser supports transitions", function () {
                var style = ".item1-one",
                    duration;
                duration = CSS.getTransDuration(style);
                expect(duration).toBe(1000);

                style = ".item2-two";
                duration = CSS.getTransDuration(style);
                expect(duration).toBe(1500);

                style = ".two-properties-transitions";
                duration = CSS.getTransDuration(style);
                expect(duration).toBe(3000);
            });
        }

        //simulating a browser without transition support
        it("CSS.getTransDuration should return 0 if the browser doesn't support transitions", function () {
            var cache = Backbone.States._CSS.transitions;
            Backbone.States._CSS.transitions = false;

            var style = ".item1-one",
                duration;
            duration = CSS.getTransDuration(style);
            expect(duration).toBe(0);

            Backbone.States._CSS.transitions = cache;
        });


    });


    describe("ViewStateTarget visibility tests", function () {

        var $el = $(createMockHtmlContainer()).find("#item1");
        var target = new Backbone.States._ViewStateTarget($el, false);

        it("after toggle called, view state target should be considered as not displayed", function () {
            target.toggle();
            expect(target.$el.css(target.display.type)).toBe(target.display.no);
        });

        it("after toggle called another time,view state target should be considered as displayed", function () {
            target.toggle();
            expect(target.$el.css(target.display.type)).toBe(target.display.yes);
        });

    });


    describe("ViewStateTarget style swapping", function () {

        var $el = $(createMockHtmlContainer()).find("#item1");
        var target = new Backbone.States._ViewStateTarget($el, {
            "display":"visibility",
            yes:"visible",
            no:"hidden"
        });

        it("should have the right style when state style applied with default css modifier", function () {
            target.addStateStyle("one");
            expect(target.$el.hasClass("item1-one")).toBeTruthy();
            target.removeStateStyle("one");
        });

        it("should not apply style that doesn't exist in any style sheet", function () {
            target.addStateStyle("no-exist");
            expect(target.$el.hasClass("no-exist")).toBeFalsy();
        });

        it("should have the right style when state style applied with custom style name modifier", function () {
            Backbone.States.config.cssModifier = "--";

            target.addStateStyle("one");
            expect(target.$el.hasClass("item1--one")).toBeTruthy();
            target.removeStateStyle("one");

            Backbone.States.config.cssModifier = "-";
        });


    });


    describe("StatesContext object builder tests from DOM", function () {

        var $el = $(createMockHtmlContainer());
        var statesContext = Backbone.States._buildStatesContext($el, {
            keepLayout:true,
            deepFetch:false,
            includeRoot:false,
            initialState:"two",
            reverseTransitions:false,
            silent:false
        });

        it("should build states names correctly", function () {
            expect(statesContext).not.toBe(null);
            expect(statesContext.names).toEqual("one,two,three");
        });


        it("each context property should be initialized with the value declared in options", function () {
            expect(statesContext.keepLayout).toBeTruthy();
            expect(statesContext.deepFetch).toBeFalsy();
            expect(statesContext.reverseTransitions).toBeFalsy();
            expect(statesContext.silent).toBeFalsy();
            expect(statesContext.initialState).toEqual("two");
        });

        it("the statesContext container property should be be $el ", function () {
            expect(statesContext.container).toEqual($el);
        });

        it("should find 3 targets in stateContext", function () {
            expect(statesContext.targets.length).toEqual(3);
        });

        it("should find 4 targets in stateContext if deepFetch if set to true", function () {
            statesContext = Backbone.States._buildStatesContext($el, {
                deepFetch:true
            });
            expect(statesContext.targets.length).toEqual(4);
        });


        it("each required context property should be initialized with the a default value when not declared in options", function () {
            statesContext = Backbone.States._buildStatesContext($el, {});

            expect(statesContext.keepLayout).toBeTruthy();
            expect(statesContext.deepFetch).toBeFalsy();
            expect(statesContext.initialState).toEqual("one");

            expect(statesContext.reverseTransitions).toBeUndefined();
            expect(statesContext.silent).toBeUndefined();

        });


    });


    describe("StatesContext object builder tests from options", function () {

        var $el = $(createMockHtmlContainer());
        var statesContext = Backbone.States._buildStatesContext($el, {
            names:"one,two,three",
            targets:{
                ".item1":{
                    inc:"one,two,three"
                },
                ".item2":{
                    inc:"one,two"
                },
                ".item3":{
                    exc:"one"
                }
            }
        });

        it("should find states names", function () {
            expect(statesContext).not.toBe(null);
            expect(statesContext.names).toEqual("one,two,three");
        });

        it("the statesContext container property should be $el object ", function () {
            expect(statesContext.container).toEqual($el);
        });


        it("keepLayout should be true by default", function () {
            expect(statesContext.keepLayout).toBeTruthy();
        });


        it("should find 3 targets in stateContext", function () {
            expect(statesContext.targets.length).toEqual(3);
        });

    });


    describe("ViewStates behaviours tests", function () {

        var $el = $(createMockHtmlContainer());
        var statesContext = Backbone.States._buildStatesContext($el, {});
        var states = Backbone.States._buildStates(statesContext);

        var stateOne = states.one;
        stateOne.enter();

        it("excluded elements should not visible after entered state", function () {
            var element;
            for (var i = 0, len = stateOne.exclusions; i < len; i++) {
                element = stateOne.exclusions[i];
                expect(element.isDisplayed()).toBe(false);
            }

        });

        it("included elements should be visible after entered state", function () {
            for (var i = 0, el, len = stateOne.inclusions.length; i < len; i++) {
                el = stateOne.inclusions[i];
                expect(el.isDisplayed()).toBe(true);
            }
        });


    });

    describe("view states builder object tests", function () {
        var $el = $(createMockHtmlContainer());
        var statesContext = Backbone.States._buildStatesContext($el, {includeRoot:true});
        var statesObject = Backbone.States._buildStates(statesContext);
        var aState, found;

        it("should find the $el object in every ViewState when includeRoot is set to true", function () {
            _.each(_.keys(statesObject), function (key) {
                aState = statesObject[key];
                found = _.find(aState.inclusions, function (viewState) {
                    return viewState.$el === $el;
                });
            });
            expect(found).toBeTruthy();
        });


    });


    describe("ViewStatesManager tests on a container with child nodes that have no transitions in class", function () {

        var mockStatesContainer = "<div id='container-test' data-states='one,two,three'>";
        mockStatesContainer = mockStatesContainer.concat("<div id='item1' class='test1' data-states-inc='one'></div>");
        mockStatesContainer = mockStatesContainer.concat("<div id='item2' class='test2' data-states-inc='two'></div>");
        mockStatesContainer = mockStatesContainer.concat("<div id='item3' class='test3' data-states-exc='three'></div>");
        mockStatesContainer = mockStatesContainer.concat("</div>");


        var $el = $(mockStatesContainer);

        var statesContext = Backbone.States._buildStatesContext($el, {});
        var states = Backbone.States._buildStates(statesContext);
        var CSS = Backbone.States._CSS;

        var mockView = {};
        _.extend(mockView, Backbone.Events);

        var statesManager = new Backbone.States._createStateManager({
            view:mockView,
            states:states
        }, {});

        it("should have a StatesTransitionManager type object when browser supports transitions", function () {
            expect(statesManager instanceof Backbone.States._StatesTransitionsManager).toBe(CSS.transitions !== false);
        });


        it("should throw error when setting unknown state", function () {
            try {
                statesManager.setState("no-exist");
            } catch (e) {
                expect(e.message).toEqual("unknown state:-no-exist-");
            }

        });

        it("Should trigger states life cycle events if not silent", function () {
            statesManager.silent = false;
            statesManager.setState("one");
            spyOn(mockView, 'trigger');

            statesManager.setState("two");
            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.EXIT, Backbone.States.EXIT, "one");
            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.EXITED, Backbone.States.EXITED, "one");
            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.ENTER, Backbone.States.ENTER, "two");
            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.CHANGE, Backbone.States.CHANGE, "two");

            statesManager.setState("three");

            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.EXIT, Backbone.States.EXIT, "two");
            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.EXITED, Backbone.States.EXITED, "two");
            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.ENTER, Backbone.States.ENTER, "three");
            expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.CHANGE, Backbone.States.CHANGE, "three");

        });

        it("Should not trigger states life cycle events if silent", function () {
            statesManager.silent = true;
            statesManager.setState("one");
            spyOn(mockView, 'trigger');

            statesManager.setState("two");
            expect(mockView.trigger).not.toHaveBeenCalledWith(Backbone.States.EXIT, Backbone.States.EXIT, "one");
            expect(mockView.trigger).not.toHaveBeenCalledWith(Backbone.States.ENTER, Backbone.States.ENTER, "two");


        });

    });

    if (Backbone.States._CSS.transitions !== false) {

        describe("ViewStatesTransitionsManager tests", function () {

            it("view should have triggered the right events after transitions end", function () {

                var node = "<div id='container-test' class='container' data-states='one,two,three'>";
                node = node.concat("<div id='item1' class='item1' data-states-inc='all'></div>");
                node = node.concat("</div>");

                var $el = $(node);
                var statesContext = Backbone.States._buildStatesContext($el, {});
                var states = Backbone.States._buildStates(statesContext);

                var mockView = {};
                _.extend(mockView, Backbone.Events);

                var statesManager = new Backbone.States._createStateManager({
                    view:mockView,
                    states:states
                }, {});

                statesManager.silent = false;
                spyOn(mockView, 'trigger');

                var flag;
                // first transition 1s
                runs(function () {
                    statesManager.setState("one");

                    setTimeout(function () {
                        flag = true;
                    }, 1000);
                });

                waitsFor(function () {
                    return flag;
                }, "The Value should be incremented", 1100);

                runs(function () {
                    expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.ENTER, Backbone.States.ENTER, "one");
                    expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.CHANGE, Backbone.States.CHANGE, "one");
                });

                // second transitions with delay
                runs(function () {
                    flag=false;
                    statesManager.setState("three");

                    setTimeout(function () {
                        flag = true;
                    }, 4000);
                });

                waitsFor(function () {
                    return flag;
                }, "The Value should be incremented", 4100);

                runs(function () {
                    expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.EXIT, Backbone.States.EXIT, "one");
                    expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.EXITED, Backbone.States.EXITED, "one");
                    expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.ENTER, Backbone.States.ENTER, "three");
                    expect(mockView.trigger).toHaveBeenCalledWith(Backbone.States.CHANGE, Backbone.States.CHANGE, "three");
                });

            });

        });
    }


}(window));
